import os


class TemplateSet(object):

    def __init__(self, root_directory):
        self.root_directory = root_directory

    def get(self, template_name):
        template_path = os.path.join(self.root_directory,
                                     template_name)

        assert os.path.exists(template_path), "cannot find template {0} at {1}".format(
            template_name,
            template_path
        )
        assert not os.path.isdir(template_path), "template path {0} is a directory".format(template_path)

        with open(template_path, 'r') as template_file:
            return template_file.read()

    def __getitem__(self, name):
        return self.get(name)



import os
import re
import yaml
from .exceptions import MiuraException
import logging

logger = logging.getLogger(__name__)


def load_data_from_path(path):
    file_paths = load_file_or_directory(path)
    return retrieve_data(file_paths)


def load_file_or_directory(path):
    """
    given a path, determine if the path is a file or directory, and
    yield a list of absolute file paths
    """
    assert os.path.exists(path), "{0} does not exist!".format(path)
    absolute_path = os.path.abspath(path)
    if not os.path.isdir(path):
        yield absolute_path
    else:
        for root, dirs, file_paths in os.walk(path):
            for file_path in file_paths:
                yield os.path.join(root, file_path)


def retrieve_data(file_paths):

    """
    passed an iterable list of file_paths, loop through all of them and
    generate a dictionary containing all the context
    """
    data_dict = {}
    for file_path in file_paths:
        with open(file_path) as fh:
            try:
                content = yaml.load(fh.read())
            except yaml.YAMLError as e:
                raise MiuraException(
                    "Unable to parse yaml at {0}: \n {1}".format(
                        file_path,
                        str(e)
                    ))
            if not isinstance(content, dict):
                raise MiuraException(
                    "{0} is does not translate to a dictionary!".format(file_path)
                )
            data_dict.update(content)
    return data_dict


def filter_data(data, filter_dict):
    """ filter a data dictionary for values only matching the filter """
    for key, match_string in filter_dict.items():
        if key not in data:
            logger.warning("{0} doesn't match a top level key".format(key))
            continue
        values = data[key]
        matcher = re.compile(match_string)
        if isinstance(values, list):
            values = [v for v in values if matcher.search(v)]
        elif isinstance(values, dict):
            values = dict((k, v) for k, v in values.items() if matcher.search(k))
        else:
            raise MiuraException("cannot filter a {0}".format(type(values)))
        data[key] = values



import os
import jinja2


def get_method_from_module(module_path, method_name):
    """ from a valid python module path, get the run method name passed """
    top_module = __import__(module_path)

    module = top_module
    # we tunnel down until we find the module we want
    for submodule_name in module_path.split('.')[1:]:
        module = getattr(module, submodule_name)

    assert hasattr(module, method_name), \
        "unable to find method {0} from module {1}. does the method exist?".format(method_name, module_path)
    return getattr(module, method_name)


def format_path_to_module(file_path):
    file_path = file_path.strip(" \n\t " + os.sep + os.curdir)
    return file_path.replace(os.sep, '.')


def specialize_content(content, data):
    """
    return a string from a jinja2 template <content> rendered with
    <data>
    """
    return jinja2.Template(content).render(**data)



"""Miura, a jenkins job management tool

Usage:
  miura docs
  miura [-d --dry-run -p <print_dir>] [-f <filter>...] <script_name> [<args>...]

Options:
  -h, --help                              Show this usage guide.
  -f <filter>..., --filter <filter>...    Filters
  -p <print_dir>, --print <print_dir>     Print the jobs out to a directory
  -d, --delete                            if set, will delete jobs
  --dry-run                               don't actually perform the operation
"""
from docopt import docopt
import logging
import os
import signal
import sys
from miura.exceptions import MiuraException

DATA_DIRECTORY = os.path.join(os.curdir, 'data')
SCRIPTS_DIRECTORY = os.path.join(os.curdir, 'scripts')
TEMPLATE_DIRECTORY = os.path.join(os.curdir, 'templates')

LOGGER = logging.getLogger(__name__)

from .script import MiuraScript


def signal_handler(signal, frame):
    print("\nShutting down miura...")
    sys.exit(0)


def main(argv=sys.argv[1:]):
    # add current directory to system path, if it's not on there already
    # this combats against buildout sandboxing sys.path
    if os.curdir not in sys.path:
        sys.path.append(os.curdir)
    signal.signal(signal.SIGINT, signal_handler)
    _create_stdout_logger()
    options = docopt(__doc__, argv=argv, options_first=True)
    try:
        filters = {}
        for filter_string in options['--filter']:
            key, value = _parse_filter_string(filter_string)
            filters[key] = value

        miura_script = MiuraScript(options['<script_name>'],
                                   DATA_DIRECTORY,
                                   SCRIPTS_DIRECTORY,
                                   TEMPLATE_DIRECTORY,
                                   data_filters=filters)
        if options['--delete']:
            miura_script.delete = True
        if options['--dry-run']:
            miura_script.dry_run = True
        if options['--print']:
            miura_script.print_directory = options['<print_dir>']
        miura_script()

    except (MiuraException, AssertionError):
        LOGGER.exception("")


def _parse_filter_string(filter_string):
    """ parse a filter string into a key-value pair """
    assert "=" in filter_string, "filter string requires an '=', got {0}".format(filter_string)
    split_values = filter_string.split('=')
    assert len(split_values) == 2, "more than one equals found in filter string {0}!".format(filter_string)
    return split_values


def _create_stdout_logger():
    """ create a logger to stdout """
    log = logging.getLogger(__name__)
    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter('%(message)s'))
    out_hdlr.setLevel(logging.INFO)
    log.addHandler(out_hdlr)
    log.setLevel(logging.INFO)



import os

from miura import runner
from .utils import get_method_from_module, format_path_to_module
from .data import load_data_from_path, filter_data
from .template import TemplateSet
from .exceptions import MiuraException

import logging

DEFAULT_DATA_DIRECTORY = os.path.join(os.curdir, 'data')
DEFAULT_TEMPLATE_DIRECTORY = os.path.join(os.curdir, 'templates')
DEFAULT_SCRIPTS_DIRECTORY = os.path.join(os.curdir, 'scripts')

LOGGER = logging.getLogger(__name__)


class MiuraScript(object):

    delete = False  # if true, delete the jobs instead of upserting them
    dry_run = False  # if true, the jobs will not run
    print_dir = None  # if set, will print configs to the directory passed

    def __init__(self, script_name,
                 data_directory,
                 scripts_directory,
                 template_directory,
                 data_filters=None
                 ):
        self.script_name = script_name
        self.data_directory = data_directory
        self.scripts_directory = scripts_directory
        self.template_directory = template_directory
        self.data_filters = data_filters or {}
        self.method_options = {}

    def __call__(self):
        target_module = "{0}.{1}".format(
            format_path_to_module(self.scripts_directory),
            self.script_name
        )

        try:
            run_method = get_method_from_module(target_module, 'run')
        except ImportError:
            raise MiuraException("Unable to find script {0}".format(target_module))

        data = load_data_from_path(self.data_directory)
        if self.data_filters:
            filter_data(data, self.data_filters)

        templates = TemplateSet(self.template_directory)

        if self.delete:
            target_method = 'delete'
        else:
            target_method = 'upsert'

        if self.dry_run:
            LOGGER.info("Performing a Dry Run! No Jobs Are Being Created")
            target_method = 'dry_run'

        job_parser = runner.JobParser(
            data,
            templates,
        )

        for job in job_parser.parse_job(run_method, self.method_options):
            if self.print_dir:
                job.print_job(self.print_dir)
            getattr(job, target_method)()



import os
import logging
from utils import specialize_content
from jenkinsapi import jenkins

LOGGER = logging.getLogger(__name__)


class JenkinsApiCache(dict):

    def __missing__(self, key):
        self[key] = jenkins.Jenkins(key)
        return self[key]


class JobParser(object):

    def __init__(self, data, templates):
        self._data = data
        self._templates = templates
        self._jenkinsapi_cache = JenkinsApiCache()

    def parse_job(self, run_method, options):
        """
        Generates and returns a job object with the following:

        * a run method, as defined in the readme
        * a list of posix-like arguments
        * a dictionary of data
        * templates: a dict-like interface of (template_name, template_body) pairs
        """

        for job_dict in run_method(options, self._data):

            # unpackaging dict
            host = job_dict['host']
            name = job_dict['name']
            template = job_dict['template']
            job_data = {
                'name': name,
                'host': host,
                'job_data': job_dict['job_data']
            }

            template_body = self._templates[template]
            config_xml = specialize_content(template_body, job_data)

            yield MiuraJenkinsJob(
                self._jenkinsapi_cache[host],
                name,
                config_xml
            )


class MiuraJenkinsJob(object):
    """
    A representation of a jenkins job in miura

    jenkins_host: jenkinsapi.Jenkins instance
    name: name of the job
    config_xml: body of the config.xml
    """

    def __init__(self, jenkins_host, name, config_xml):
        self.jenkins_host = jenkins_host
        self.name = name
        self.config_xml = config_xml

    def upsert(self):
        """ create or update the jenkins job """
        if not self.jenkins_host.has_job(self.name):
            LOGGER.info("creating {0}...".format(self.name))
            self.jenkins_host.create_job(self.name, self.config_xml)
        else:
            jenkins_job = self.jenkins_host[self.name]
            LOGGER.info("updating {0}...".format(self.name))
            jenkins_job.update_config(self.config_xml)

    def delete(self):
        """ delete the jenkins job, if it exists """
        if self.jenkins_host.has_job(self.name):
            LOGGER.info("deleting {0}...".format(self.name))
            self.jenkins_host.delete_job(self.name)

    def print_job(self, directory):
        if not os.path.exists(directory):
            os.makedirs(directory)

        job_filename = "{0}-{1}.xml".format(
            self.jenkins_host.baseurl, self.name
        )

        with open(os.path.join(directory, job_filename), 'w+') as fh:
            fh.write(self.config_xml)

    def dry_run(self):
        """ print information about the jenkins job """
        LOGGER.info("Job Info: {name} -> {host}".format(
            name=self.name,
            host=self.jenkins_host.baseurl
        ))
